<?php

/**
 * @file
 * Contains CdekSerializer class.
 */

/**
 * CDEK API serializer.
 */
class CdekSerializer {

  /**
   * Ignored node types for XML format.
   *
   * @var array
   */
  private static $xmlIgnoredNodeTypes = array(XML_PI_NODE, XML_COMMENT_NODE);

  /**
   * Decodes data from the JSON format.
   *
   * @param string $data
   *   The encoded JSON string to decode.
   *
   * @return mixed
   *   The decoded data.
   *
   * @throws UnexpectedValueException
   *   Error while decoding JSON.
   */
  public static function decodeJson($data) {
    $decoded_data = json_decode($data, TRUE);
    if (json_last_error() !== JSON_ERROR_NONE) {
      throw new UnexpectedValueException(json_last_error_msg());
    }
    return $decoded_data;
  }

  /**
   * Decodes data from the XML format.
   *
   * @param string $data
   *   The encoded XML string to decode.
   *
   * @return mixed
   *   The decoded data.
   *
   * @throws UnexpectedValueException
   *   Error while decoding XML.
   */
  public static function decodeXml($data) {
    if (trim($data) === '') {
      throw new UnexpectedValueException('Invalid XML data, it can not be empty.');
    }

    // Enable libxml errors and the ability to load external entities.
    $internal_errors = libxml_use_internal_errors(TRUE);
    $disable_entities = libxml_disable_entity_loader(TRUE);
    libxml_clear_errors();
    // Load XML from a string.
    $dom = new DOMDocument();
    $dom->loadXML($data, LIBXML_NONET | LIBXML_NOBLANKS);
    // Return the previous settings.
    libxml_use_internal_errors($internal_errors);
    libxml_disable_entity_loader($disable_entities);

    // Check last error from libxml.
    if ($error = libxml_get_last_error()) {
      libxml_clear_errors();
      throw new UnexpectedValueException($error->message);
    }

    // Get the root node.
    $root_node = NULL;
    foreach ($dom->childNodes as $child) {
      if ($child->nodeType === XML_DOCUMENT_TYPE_NODE) {
        throw new UnexpectedValueException('Document types are not allowed.');
      }
      if (!$root_node && !in_array($child->nodeType, self::$xmlIgnoredNodeTypes, TRUE)) {
        $root_node = $child;
      }
    }
    // Check the existence of the root node.
    if ($root_node === NULL) {
      throw new UnexpectedValueException('The root node name is not correctly configured.');
    }

    /** @var DOMNode $root_node */
    // Parse the children of the root node.
    if ($root_node->hasChildNodes()) {
      return self::parseXml($root_node);
    }
    // Returns only root node value.
    if (!$root_node->hasAttributes()) {
      return $root_node->nodeValue;
    }

    // Return only the array of root node attributes.
    $data = array();
    foreach ($root_node->attributes as $attribute) {
      $data[$attribute->nodeName] = $attribute->nodeValue;
    }
    return $data;
  }

  /**
   * Parse the input DOMNode into an array or a string.
   *
   * @param DOMNode $node
   *   The DOMNode for parsing.
   *
   * @return array|string
   *   The result of parsing.
   */
  private static function parseXml(DOMNode $node) {
    $data = self::parseXmlAttributes($node);
    $value = self::parseXmlValue($node);

    // Attributes are not present, return only value.
    if (!count($data)) {
      return $value;
    }
    // The value is not an array, return an array of attributes.
    if (!is_array($value)) {
      return $data;
    }
    // Value is an array, combine it with an array of attributes.
    foreach ($value as $key => $val) {
      $data[$key] = $val;
    }
    return $data;
  }

  /**
   * Parse the input DOMNode attributes into an array.
   *
   * @param DOMNode $node
   *   The DOMNode for parsing.
   *
   * @return array
   *   The result of parsing.
   */
  private static function parseXmlAttributes(DOMNode $node) {
    if (!$node->hasAttributes()) {
      return array();
    }

    $data = array();
    foreach ($node->attributes as $attribute) {
      $data[$attribute->nodeName] = $attribute->nodeValue;
    }
    return $data;
  }

  /**
   * Parse the input DOMNode value into an array or a string.
   *
   * @param DOMNode $node
   *   The DOMNode for parsing.
   *
   * @return array|string
   *   The result of parsing.
   */
  private static function parseXmlValue(DOMNode $node) {
    if (!$node->hasChildNodes()) {
      return $node->nodeValue;
    }
    // Check the lone child node of a special type.
    $node_types = array(XML_TEXT_NODE, XML_CDATA_SECTION_NODE);
    if ($node->childNodes->length === 1 && in_array($node->firstChild->nodeType, $node_types)) {
      return $node->firstChild->nodeValue;
    }

    $value = array();
    foreach ($node->childNodes as $sub_node) {
      if (in_array($sub_node->nodeType, self::$xmlIgnoredNodeTypes, TRUE)) {
        continue;
      }
      // Parse node child.
      $val = self::parseXml($sub_node);
      $value[$sub_node->nodeName][] = $val;
    }
    // Simplify the arrays with one element.
    foreach ($value as $key => $val) {
      if (is_array($val) && count($val) === 1) {
        $value[$key] = current($val);
      }
    }
    return $value;
  }

}
